# Script to shift points to lines.
# Copyright (C) 2016 by Igor Belov
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Parameters for Processing algorithm
##Point_layer=vector
##Point_field=field Point_layer
##Line_layer=vector
##Line_field=field Line_layer
##Ignore_fields=boolean False
##Output_layer=output vector

from processing.core.GeoAlgorithmExecutionException import GeoAlgorithmExecutionException
from qgis.core import (
    QGis,
    QgsField,
    QgsGeometry,
    QgsPoint,
    QgsCoordinateReferenceSystem,
    QgsVectorFileWriter
)
from PyQt4.QtCore import QVariant
import math

BIGNUM = 1.7976931348623157e+308

lin_layer = processing.getObject(Line_layer)
if lin_layer.wkbType() != QGis.WKBLineString:
    raise GeoAlgorithmExecutionException('Line layer must be LineString')

lin_fldnum = 0
for field in lin_layer.pendingFields():
    if Line_field == field.name():
        lin_fldtype = field.typeName()
        break
    lin_fldnum += 1

pt_layer = processing.getObject(Point_layer)
if pt_layer.wkbType() != QGis.WKBPoint:
    raise GeoAlgorithmExecutionException('Point layer must be Point')

pt_crs = pt_layer.dataProvider().crs()
fields = pt_layer.pendingFields()
pt_fldnum = 0
for field in fields:
    if Point_field == field.name():
        pt_fldtype = field.typeName()
        break
    pt_fldnum += 1

if lin_fldtype != pt_fldtype and not Ignore_fields:
    raise GeoAlgorithmExecutionException('Line and Point field types must coincide')

fieldnames = [field.name() for field in fields]
for fldname in ['ucoord', 'vcoord']:
    if fldname in fieldnames:
        i = 1
        while True:
            fldname_i = '{:s}_{:d}'.format(fldname, i)
            if fldname_i in fieldnames:
                i += 1
            else:
                fldname = fldname_i
                break

    fields.append(QgsField(fldname, QVariant.Double, "Real", 23, 15, ""))

writer = QgsVectorFileWriter(Output_layer, 'UTF-8', fields, QGis.WKBPoint, pt_crs, 'ESRI Shapefile')
if writer.hasError() != QgsVectorFileWriter.NoError:
    raise GeoAlgorithmExecutionException('Error creating shapefile')

lin_f = []
lin_vtx = []
iter = lin_layer.getFeatures()
for f in iter:
    g = []
    points = f.geometry().asPolyline()
    x, y = BIGNUM, BIGNUM
    for pt in points:
        if pt.x() != x or pt.y() != y:
            x, y = pt.x(), pt.y()
            g.append([x, y])

    lin_f.append(f.attributes()[lin_fldnum])
    lin_vtx.append(g)

lin_seg = []
for i in range(len(lin_vtx)):
    lin_vtx[i][0].append(0.0)
    g = []
    for j in range(len(lin_vtx[i]) - 1):
        dx = lin_vtx[i][j+1][0] - lin_vtx[i][j][0]
        dy = lin_vtx[i][j+1][1] - lin_vtx[i][j][1]
        d = math.hypot(dx, dy)
        g.append([dx/d, dy/d, d])
        lin_vtx[i][j+1].append(lin_vtx[i][j][2] + d)

    lin_seg.append(g)

iter = pt_layer.getFeatures()
total = 100.0 / pt_layer.featureCount()
current = 0
for f in iter:
    progress.setPercentage(int(current * total))
    pt = f.geometry().asPoint()
    x, y = pt.x(), pt.y()
    attrs = f.attributes()
    v = BIGNUM
    imin = -1
    for i in range(len(lin_f)):
        if Ignore_fields or attrs[pt_fldnum] == lin_f[i]:
            for j in range(len(lin_seg[i])):
                dx = x - lin_vtx[i][j][0]
                dy = y - lin_vtx[i][j][1]
                uj = lin_seg[i][j][0] * dx + lin_seg[i][j][1] * dy
                vj = lin_seg[i][j][0] * dy - lin_seg[i][j][1] * dx
                if j > 0 and uj < 0.0:
                    uj = 0.0
                    vj = math.copysign(math.hypot(dx, dy), vj)
                elif j < len(lin_seg[i]) - 1 and uj > lin_seg[i][j][2]:
                    continue

                if abs(vj) < abs(v):
                    imin = i
                    jmin = j
                    u, v = uj, vj

    if imin >= 0:
        x = lin_vtx[imin][jmin][0] + lin_seg[imin][jmin][0] * u
        y = lin_vtx[imin][jmin][1] + lin_seg[imin][jmin][1] * u
        attrs.extend([lin_vtx[imin][jmin][2] + u, v])

    f.setGeometry(QgsGeometry.fromPoint(QgsPoint(x, y)))
    f.setAttributes(attrs)
    writer.addFeature(f)
    current += 1

del writer
